import numpy as np
import galois

flag = "inno{h4rdc0d1ng_15_b4d}"
intermediate_flag = "inno{1nv3rt3d_m4tr1x_c00l}"

n = 127
k = 32
order = 2 ** 7
GF = galois.GF(order)


def get_random_permutation_matrix():
    p = [48, 28, 97, 79, 63, 82, 22, 46, 56, 5, 45, 37, 95, 110, 118, 31, 24, 120, 83, 115, 61, 25, 18, 101, 19, 96, 70, 81, 
         53, 41, 8, 73, 44, 26, 85, 62, 7, 16, 27, 35, 6, 99, 76, 88, 39, 89, 34, 103, 121, 2, 87, 125, 90, 122, 104, 93, 65, 
         84, 4, 106, 68, 14, 51, 57, 20, 50, 54, 107, 43, 38, 100, 29, 3, 10, 12, 13, 109, 1, 111, 49, 123, 126, 80, 32, 75, 
         71, 114, 91, 58, 74, 9, 17, 59, 40, 113, 77, 23, 69, 42, 30, 0, 21, 78, 124, 66, 52, 117, 116, 11, 102, 98, 33, 60, 
         108, 55, 72, 112, 36, 94, 86, 64, 119, 67, 92, 15, 105, 47] # totally random permutation :)
    P = GF.Zeros((n, n))
    for i in range(n):
        P[i, p[i]] = 1
    return P


def get_nonsingular_matrix():
    S = GF.Random((k, k))
    while np.linalg.det(S) == 0:
        S = GF.Random((k, k))
    return S


def pad_message(msg: bytes, pad_size: int) -> list[int]:
    padding = pad_size - (len(msg) % pad_size)
    return list(msg + padding.to_bytes() * padding)


def unpad_message(msg):
    padding_byte = msg[-1]
    for i in range(1, padding_byte + 1):
        if msg[-i] != padding_byte:
            raise ValueError("Incorrect padding!")
    return msg[:-padding_byte]


def generate_pubkey():
    rs = galois.ReedSolomon(n, k, field=GF)
    S = get_nonsingular_matrix()
    G = rs.G 
    P = get_random_permutation_matrix()

    # @ это оператор умножения матриц в numpy
    G_ = S @ G @ P

    return G_


# Экспортируем публичный ключ в удобоваримый формат
def export_pubkey(G_):
    rows = [", ".join([str(int(cell)) for cell in row]) for row in G_]
    output = "G_ = [ " + ",\n".join([f"[{row}]" for row in rows]) + " ]"

    with open("pubkey.py", "w") as f:
        f.write(output)


def encrypt(G_, flag):
    msg = pad_message(flag.encode(), k)
    
    # не забываем, что работаем в поле Галуа
    m = GF(msg)
    
    # это то же самое, что и rs.encode(m)
    c = m.T @ G_
    return c

def encrypt_intermediate_flag(message):
    msg = pad_message(message.encode(), n)

    m = GF(msg)
    
    P = get_random_permutation_matrix()
    c = m @ P

    return c

if __name__ == "__main__":
    # генрируем ключ
    G_ = generate_pubkey()
    
    # экспортируем его
    export_pubkey(G_)
    
    intermediate_ct = encrypt_intermediate_flag(intermediate_flag)
    print(f"Encrypted intermediate flag: {list(map(int, intermediate_ct))}")

    ct = encrypt(G_, flag)
    print(f"Encrypted flag: {list(map(int, ct))}")

# Чтобы декодировать кодовое слово Рида-Соломона, настоятельно рекомендуется использовать функцию galois.ReedSolomon.decode

# Вывод программы:
# Encrypted intermediate flag: [101, 101, 101, 101, 101, 114, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 48, 108, 101, 101, 110, 101, 116, 99, 101, 101, 110, 101, 101, 52, 101, 101, 101, 101, 101, 51, 101, 101, 101, 101, 101, 101, 101, 116, 118, 101, 105, 101, 101, 101, 101, 101, 101, 101, 51, 101, 101, 101, 101, 95, 101, 123, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 111, 101, 101, 49, 49, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 101, 100, 125, 110, 101, 101, 101, 48, 101, 101, 101, 101, 101, 101, 101, 101, 95, 101, 101, 101, 101, 120, 101, 101, 109, 101, 114, 101, 101, 101, 101, 101, 101]
# Encrypted flag: [97, 73, 26, 65, 127, 121, 101, 109, 102, 47, 45, 116, 61, 102, 9, 8, 21, 17, 32, 57, 86, 33, 45, 104, 74, 96, 98, 109, 95, 95, 87, 7, 46, 62, 61, 16, 126, 95, 81, 40, 122, 124, 77, 126, 112, 22, 58, 112, 26, 23, 114, 110, 13, 88, 105, 37, 33, 3, 125, 118, 78, 71, 50, 35, 15, 46, 24, 56, 40, 20, 121, 119, 57, 18, 89, 114, 63, 5, 103, 46, 69, 9, 8, 85, 115, 43, 92, 126, 32, 116, 96, 104, 100, 47, 25, 110, 70, 33, 13, 85, 105, 39, 97, 3, 93, 26, 69, 121, 106, 63, 81, 39, 56, 62, 119, 24, 120, 79, 9, 65, 48, 64, 55, 80, 8, 69, 45]